;;; Gash --- Guile As SHell
;;; Copyright © 2018 Jan (janneke) Nieuwenhuizen <janneke@gnu.org>
;;;
;;; This file is part of Gash.
;;;
;;; Gash is free software; you can redistribute it and/or modify it
;;; under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 3 of the License, or (at
;;; your option) any later version.
;;;
;;; Gash is distributed in the hope that it will be useful, but WITHOUT ANY
;;; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;;; details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with Gash.  If not, see <http://www.gnu.org/licenses/>.

(define-module (gash builtins)
  #:use-module (ice-9 ftw)
  #:use-module (ice-9 getopt-long)
  #:use-module (ice-9 local-eval)
  #:use-module (ice-9 match)
  #:use-module (ice-9 pretty-print)
  #:use-module (ice-9 receive)
  #:use-module (ice-9 regex)

  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-26)

  #:use-module (gash config)
  #:use-module (gash gash)              ; %prefer-builtins?
  #:use-module (gash bournish-commands)
  #:use-module (gash environment)
  #:use-module (gash shell-utils)
  #:use-module (gash io)
  #:use-module (gash job)
  #:use-module (gash pipe)
  #:use-module (gash script)
  #:use-module (gash util)

  #:export (
            %builtin-commands
            PATH-search-path
            trace

            bg-command
            cd-command
            echo-command
            eval-command
            exit-command
            fg-command
            help-command
            jobs-command
            pwd-command
            set-command
            shift-command
            ))

(define (cd-command . args)
  (match args
    (() (cd-command (getenv "HOME")))
    ((dir)
     (let ((old (variable "OLDPWD")))
       (assignment "OLDPWD" (getcwd))
       (catch #t
         (lambda _
           (if (string=? dir "-") (chdir old)
               (chdir dir))
           0)
         (lambda (key command fmt args exit)
           (apply format (current-error-port) "cd: ~a: ~a\n" (cons dir args))
           1))))
    ((args ...)
     (format (current-error-port) "cd: too many arguments: ~a\n" (string-join args)))))

(define (echo-command . args)
  (lambda _
    (match args
      (() (newline))
      (("-n" args ...) (display (string-join args)))
      (_ (display (string-join args)) (newline)))))

(define (bg-command . args)
  (match args
    (() (bg 1))
    ((job x ...) (bg (string->number (car job))))))

(define (fg-command . args)
  (match args
    (() (fg 1))
    ((job x ...) (fg (string->number (car job))))))

(define (jobs-command)
  (format (current-error-port) "jobs: ~s\n" job-table)
  (for-each (lambda (job) (display-job job)) (reverse job-table)))

(define (pwd-command . _)
  (lambda _ (stdout (getcwd))))

(define (set-command . args) ;; TODO export; env vs set
  (define (display-var o)
    (format #t "~a=~a\n" (car o) (cdr o)))
  (match args
    (() (lambda _ (for-each display-var %global-variables)))
    (("-e") (set-shell-opt! "errexit" #t))
    (("+e") (set-shell-opt! "errexit" #f))
    (("-u") (set-shell-opt! "nounset" #t))
    (("+u") (set-shell-opt! "nounset" #f))
    (("-x") (set-shell-opt! "xtrace" #t))
    (("+x") (set-shell-opt! "xtrace" #f))
    (("-o" option) (format (current-error-port) "warning: set: not supported: ~a\n" args))
    (("+o" option) (format (current-error-port) "warning: set: not supported: ~a\n" args))
    (((and (? string?) arg)) (let* ((lst (string->string-list arg))
                                    (set (car lst)))
                               (when (not (member set '("-" "+")))
                                 (error (format #f "set: no such option:~s\n" args)))
                               (apply set-command (map (cut string-append set <>) (cdr lst)))))
    ((h ...) (last (map set-command args)))))

(define (shift-command . args)
  (lambda _
    (match args
      (() (when (pair? (cdr (%command-line)))
            (%command-line (cons (car (%command-line)) (cddr (%command-line)))))))))

(define (eval-command . args)
  (lambda _
    (match args
      (() #t)
      ((args ...)
       (let ((ast (parse-string (string-join args))))
         ;;(ignore-error (run ast))
         (run ast)
         (assignment "?" "0")
         #t)))))

(define (exit-command . args)
  (match args
    (() (exit 0))
    ((status)
     (exit (string->number status)))
    ((args ...)
     (format (current-error-port) "exit: too many arguments: ~a\n" (string-join args)))))

(define (help-command . _)
  (lambda _
    (display "\
Hello, this is GASH, Guile As SHell.

GASH is work in progress; many language constructs work, globbing
mostly works, pipes work, some redirections work.
")
    (display "\nIt has these builtin commands:\n")
    (display-tabulated (map car %builtin-commands))
    (when (or %prefer-builtins? (not (PATH-search-path "ls")))
      (display "\nand features the following, somewhat naive, bournish commands:\n")
      (display-tabulated (map car %bournish-commands)))))

(define (term->string o)
  (match o
    ((? string?) o)
    (('variable name) (variable name))
    (('variable-or name default) (variable-or name default))
    (('variable-and name default) (variable-and name default))
    (_ (format #f "~s" o))))

(define (trace commands)
  `(xtrace
    ,(lambda _
       (when (shell-opt? "xtrace")
         (for-each
          (lambda (o)
            (match o
              (('command (and command (or (? string?) ('variable _))) ...)
               (format (current-error-port) "+ ~a\n" (string-join (map term->string command))))
              (('command ('assignment name value))
               (format (current-error-port) "+ ~a=~a\n" name (term->string value)))
              (_ (format (current-error-port) "+ ~s <FIXME>\n" o))))
          (reverse commands))))))

(define %builtin-commands
  `(
    ("bg"      . ,bg-command)
    ("cd"      . ,cd-command)
    ("echo"    . ,echo-command)
    ("eval"    . ,eval-command)
    ("exit"    . ,exit-command)
    ("fg"      . ,fg-command)
    ("help"    . ,help-command)
    ("jobs"    . ,jobs-command)
    ("pwd"     . ,pwd-command)
    ("set"     . ,set-command)
    ("shift"   . ,shift-command)
    ))
